# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas
from mlflow.tracking.artifact_utils import _download_artifact_from_uri
from mlflow.utils.file_utils import TempDir
from pyspark.sql import SparkSession

from merlinpyspark.config import ModelConfig
from merlinpyspark.spec.prediction_job_pb2 import ModelType


def create_model_udf(spark_session: SparkSession, model_config: ModelConfig, features):
    if model_config.model_type() == ModelType.Name(ModelType.PYFUNC_V2):
        return spark_udf(spark_session, model_config.model_uri(), features,
                         result_type=model_config.result_type())

    raise ValueError(f"model type not supported: {model_config.model_type()}")


def spark_udf(spark, model_uri, features, result_type="double"):
    """
    Create spark pandas udf given the model uri
    :param spark: spark context
    :param model_uri: path to model
    :param features: list containing the feature names
    :param result_type: result type of the model
    :return:
    """
    # Scope Spark import to this method so users don't need pyspark to use non-Spark-related
    # functionality.
    from mlflow.pyfunc.spark_model_cache import SparkModelCache
    from pyspark.sql.functions import pandas_udf
    from pyspark.sql.types import _parse_datatype_string
    from pyspark.sql.types import ArrayType, DataType
    from pyspark.sql.types import DoubleType, IntegerType, FloatType, LongType, StringType

    if not isinstance(result_type, DataType):
        result_type = _parse_datatype_string(result_type)

    elem_type = result_type
    if isinstance(elem_type, ArrayType):
        elem_type = elem_type.elementType

    supported_types = [IntegerType, LongType, FloatType, DoubleType, StringType]

    if not any([isinstance(elem_type, x) for x in supported_types]):
        raise ValueError(
            "Invalid result_type '{}'. Result type can only be one of or an array of one "
            "of the following types types: {}".format(str(elem_type), str(supported_types)))

    with TempDir() as local_tmpdir:
        local_model_path = _download_artifact_from_uri(
            artifact_uri=model_uri, output_path=local_tmpdir.path())
        archive_path = SparkModelCache.add_local_model(spark, local_model_path)

    def predict(*args):
        model, _ = SparkModelCache.get_or_load(archive_path)
        schema = {features[i]: arg for i, arg in enumerate(args)}
        pdf = None
        for x in args:
            if type(x) == pandas.DataFrame:
                if len(args) != 1:
                    raise Exception("If passing a StructType column, there should be only one "
                                    "input column, but got %d" % len(args))
                pdf = x
        if pdf is None:
            pdf = pandas.DataFrame(schema)
        result = model.predict(pdf)
        if not isinstance(result, pandas.DataFrame):
            result = pandas.DataFrame(data=result)

        elif type(elem_type) == IntegerType:
            result = result.select_dtypes([np.byte, np.ubyte, np.short, np.ushort,
                                           np.int32]).astype(np.int32)

        elif type(elem_type) == LongType:
            result = result.select_dtypes([np.byte, np.ubyte, np.short, np.ushort, np.int, np.long])

        elif type(elem_type) == FloatType:
            result = result.select_dtypes(include=(np.number,)).astype(np.float32)

        elif type(elem_type) == DoubleType:
            result = result.select_dtypes(include=(np.number,)).astype(np.float64)

        if len(result.columns) == 0:
            raise ValueError(
                "The the model did not produce any values compatible with the requested "
                "type '{}'. Consider requesting udf with StringType or "
                "Arraytype(StringType).".format(str(elem_type)))

        if type(elem_type) == StringType:
            result = result.applymap(str)

        if type(result_type) == ArrayType:
            return pandas.Series([row[1].values for row in result.iterrows()])
        else:
            return result[result.columns[0]]

    return pandas_udf(predict, result_type)



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod

from pyspark.sql import DataFrame

from merlinpyspark.config import SinkConfig, BigQuerySinkConfig


def create_sink(sink_config: SinkConfig) -> "Sink":
    if sink_config.sink_type() == BigQuerySinkConfig.TYPE:
        if not isinstance(sink_config, BigQuerySinkConfig):
            raise ValueError("sink_config is not BigQuerySink")

        return BigQuerySink(sink_config)

    raise ValueError(f"sink type is not implemented: {sink_config.sink_type()}")


class Sink(ABC):
    @abstractmethod
    def save(self, df):
        pass


class BigQuerySink(Sink):
    WRITE_FORMAT = "bigquery"
    OPTION_TABLE = "table"
    OPTION_STAGING_BUCKET = "temporaryGcsBucket"

    def __init__(self, config: BigQuerySinkConfig):
        self._config = config

    def save(self, df: DataFrame):
        df.write \
            .mode(self._config.save_mode()) \
            .format(self.WRITE_FORMAT) \
            .option(self.OPTION_TABLE, self._config.table()) \
            .option(self.OPTION_STAGING_BUCKET, self._config.staging_bucket()) \
            .options(**self._config.options()) \
            .save()



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Iterable, MutableMapping, Dict

import yaml
from google.protobuf import json_format
from pyspark.sql.types import ArrayType, DataType, DoubleType, FloatType, \
    IntegerType, LongType, StringType

from merlinpyspark.spec.prediction_job_pb2 import BigQuerySink, BigQuerySource, Model, \
    PredictionJob, ResultType, ModelType, SaveMode


def printJobConfig(spec_path):
    with open(spec_path, "r") as f:
        print(f"""
===============================================================
Prediction job configuration:
{f.read()}
===============================================================
        """)


def load(spec_path: str) -> "JobConfig":
    """
    Read a job spec file (yaml file) and return nested dictionary
    containing the job config
    :param spec_path: path to yaml config file
    :return: JobConfig object
    """
    printJobConfig(spec_path)
    prediction_job = PredictionJob()
    with open(spec_path, "r") as f:
        return JobConfig(
            json_format.ParseDict(yaml.safe_load(f), prediction_job))


class JobConfig:
    BIGQUERY_SOURCE = "bigquery_source"
    BIGQUERY_SINK = "bigquery_sink"
    GCS_SOURCE = "gcs_source"
    GCS_SINK = "gcs_sink"

    def __init__(self, config: PredictionJob):
        self._config = config

    def source(self) -> "SourceConfig":
        src = self._config.WhichOneof("source")
        if src is None:
            raise ValueError("missing mandatory source config")

        if src == self.BIGQUERY_SOURCE:
            return BigQuerySourceConfig(self._config.bigquery_source)
        else:
            raise ValueError(f"{src} is not implemented")

    def model(self) -> "ModelConfig":
        return ModelConfig(self._config.model)

    def sink(self) -> "SinkConfig":
        sink = self._config.WhichOneof("sink")
        if sink is None:
            raise ValueError("missing mandatory sink config")

        if sink == self.BIGQUERY_SINK:
            return BigQuerySinkConfig(self._config.bigquery_sink)
        else:
            raise ValueError(f"{sink} is not implemented")


class SourceConfig(ABC):
    @abstractmethod
    def source_type(self) -> str:
        pass

    @abstractmethod
    def options(self) -> MutableMapping[str, str]:
        pass

    @abstractmethod
    def features(self) -> Iterable[str]:
        pass


class BigQuerySourceConfig(SourceConfig):
    TYPE = "bigquery"

    def __init__(self, bq_src_proto: BigQuerySource):
        self._proto = bq_src_proto

    def source_type(self) -> str:
        return self.TYPE

    def options(self) -> MutableMapping[str, str]:
        return self._proto.options

    def features(self) -> Iterable[str]:
        return self._proto.features

    def table(self) -> str:
        return self._proto.table


class ModelConfig:
    PRIMITIVE_TYPE_MAP = {
        ResultType.DOUBLE: DoubleType(),
        ResultType.FLOAT: FloatType(),
        ResultType.INTEGER: IntegerType(),
        ResultType.LONG: LongType(),
        ResultType.STRING: StringType(),
    }

    def __init__(self, model_proto: Model):
        self._proto = model_proto

    def model_type(self) -> str:
        return ModelType.Name(self._proto.type)

    def model_uri(self) -> str:
        return self._proto.uri

    def result_type(self) -> DataType:
        rt = self._proto.result.type
        if rt == ResultType.ARRAY:
            it = self._proto.result.item_type
            if it in self.PRIMITIVE_TYPE_MAP:
                return ArrayType(self.PRIMITIVE_TYPE_MAP[it])
            raise ValueError(f"unknown item type for array: {it}")

        if rt in self.PRIMITIVE_TYPE_MAP:
            return self.PRIMITIVE_TYPE_MAP[rt]

        raise ValueError(f"unknown result type: {rt}")


class SinkConfig(ABC):
    @abstractmethod
    def sink_type(self) -> str:
        pass

    @abstractmethod
    def result_column(self) -> str:
        pass

    @abstractmethod
    def options(self) -> MutableMapping[str, str]:
        pass

    @abstractmethod
    def save_mode(self) -> str:
        pass


class BigQuerySinkConfig(SinkConfig):
    TYPE = "bigquery"
    DEFAULT_RESULT_COLUMN = "prediction"
    
    def __init__(self, bq_sink_proto: BigQuerySink):
        self._proto = bq_sink_proto

    def sink_type(self) -> str:
        return self.TYPE

    def result_column(self) -> str:
        col = self._proto.result_column
        if col == "":
            return self.DEFAULT_RESULT_COLUMN
        return col

    def options(self) -> MutableMapping[str, str]:
        return self._proto.options

    def save_mode(self) -> str:
        return SaveMode.Name(self._proto.save_mode).lower()

    def table(self) -> str:
        return self._proto.table

    def staging_bucket(self) -> str:
        return self._proto.staging_bucket



# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import sys
from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
    EnumDescriptor as google___protobuf___descriptor___EnumDescriptor,
)

from google.protobuf.internal.containers import (
    RepeatedScalarFieldContainer as google___protobuf___internal___containers___RepeatedScalarFieldContainer,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)

from typing import (
    Iterable as typing___Iterable,
    List as typing___List,
    Mapping as typing___Mapping,
    MutableMapping as typing___MutableMapping,
    Optional as typing___Optional,
    Text as typing___Text,
    Tuple as typing___Tuple,
    Union as typing___Union,
    cast as typing___cast,
    overload as typing___overload,
)

from typing_extensions import (
    Literal as typing_extensions___Literal,
)


builtin___bool = bool
builtin___bytes = bytes
builtin___float = float
builtin___int = int
builtin___str = str
if sys.version_info < (3,):
    builtin___buffer = buffer
    builtin___unicode = unicode


class ResultType(builtin___int):
    DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
    @classmethod
    def Name(cls, number: builtin___int) -> builtin___str: ...
    @classmethod
    def Value(cls, name: builtin___str) -> 'ResultType': ...
    @classmethod
    def keys(cls) -> typing___List[builtin___str]: ...
    @classmethod
    def values(cls) -> typing___List['ResultType']: ...
    @classmethod
    def items(cls) -> typing___List[typing___Tuple[builtin___str, 'ResultType']]: ...
    DOUBLE = typing___cast('ResultType', 0)
    FLOAT = typing___cast('ResultType', 1)
    INTEGER = typing___cast('ResultType', 2)
    LONG = typing___cast('ResultType', 3)
    STRING = typing___cast('ResultType', 4)
    ARRAY = typing___cast('ResultType', 10)
DOUBLE = typing___cast('ResultType', 0)
FLOAT = typing___cast('ResultType', 1)
INTEGER = typing___cast('ResultType', 2)
LONG = typing___cast('ResultType', 3)
STRING = typing___cast('ResultType', 4)
ARRAY = typing___cast('ResultType', 10)

class ModelType(builtin___int):
    DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
    @classmethod
    def Name(cls, number: builtin___int) -> builtin___str: ...
    @classmethod
    def Value(cls, name: builtin___str) -> 'ModelType': ...
    @classmethod
    def keys(cls) -> typing___List[builtin___str]: ...
    @classmethod
    def values(cls) -> typing___List['ModelType']: ...
    @classmethod
    def items(cls) -> typing___List[typing___Tuple[builtin___str, 'ModelType']]: ...
    INVALID_MODEL_TYPE = typing___cast('ModelType', 0)
    XGBOOST = typing___cast('ModelType', 1)
    TENSORFLOW = typing___cast('ModelType', 2)
    SKLEARN = typing___cast('ModelType', 3)
    PYTORCH = typing___cast('ModelType', 4)
    ONNX = typing___cast('ModelType', 5)
    PYFUNC = typing___cast('ModelType', 6)
    PYFUNC_V2 = typing___cast('ModelType', 7)
INVALID_MODEL_TYPE = typing___cast('ModelType', 0)
XGBOOST = typing___cast('ModelType', 1)
TENSORFLOW = typing___cast('ModelType', 2)
SKLEARN = typing___cast('ModelType', 3)
PYTORCH = typing___cast('ModelType', 4)
ONNX = typing___cast('ModelType', 5)
PYFUNC = typing___cast('ModelType', 6)
PYFUNC_V2 = typing___cast('ModelType', 7)

class FileFormat(builtin___int):
    DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
    @classmethod
    def Name(cls, number: builtin___int) -> builtin___str: ...
    @classmethod
    def Value(cls, name: builtin___str) -> 'FileFormat': ...
    @classmethod
    def keys(cls) -> typing___List[builtin___str]: ...
    @classmethod
    def values(cls) -> typing___List['FileFormat']: ...
    @classmethod
    def items(cls) -> typing___List[typing___Tuple[builtin___str, 'FileFormat']]: ...
    INVALID_FILE_FORMAT = typing___cast('FileFormat', 0)
    CSV = typing___cast('FileFormat', 1)
    PARQUET = typing___cast('FileFormat', 2)
    AVRO = typing___cast('FileFormat', 3)
    JSON = typing___cast('FileFormat', 4)
INVALID_FILE_FORMAT = typing___cast('FileFormat', 0)
CSV = typing___cast('FileFormat', 1)
PARQUET = typing___cast('FileFormat', 2)
AVRO = typing___cast('FileFormat', 3)
JSON = typing___cast('FileFormat', 4)

class SaveMode(builtin___int):
    DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
    @classmethod
    def Name(cls, number: builtin___int) -> builtin___str: ...
    @classmethod
    def Value(cls, name: builtin___str) -> 'SaveMode': ...
    @classmethod
    def keys(cls) -> typing___List[builtin___str]: ...
    @classmethod
    def values(cls) -> typing___List['SaveMode']: ...
    @classmethod
    def items(cls) -> typing___List[typing___Tuple[builtin___str, 'SaveMode']]: ...
    ERRORIFEXISTS = typing___cast('SaveMode', 0)
    OVERWRITE = typing___cast('SaveMode', 1)
    APPEND = typing___cast('SaveMode', 2)
    IGNORE = typing___cast('SaveMode', 3)
    ERROR = typing___cast('SaveMode', 4)
ERRORIFEXISTS = typing___cast('SaveMode', 0)
OVERWRITE = typing___cast('SaveMode', 1)
APPEND = typing___cast('SaveMode', 2)
IGNORE = typing___cast('SaveMode', 3)
ERROR = typing___cast('SaveMode', 4)

class PredictionJob(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    version = ... # type: typing___Text
    kind = ... # type: typing___Text
    name = ... # type: typing___Text

    @property
    def bigquery_source(self) -> BigQuerySource: ...

    @property
    def gcs_source(self) -> GcsSource: ...

    @property
    def model(self) -> Model: ...

    @property
    def bigquery_sink(self) -> BigQuerySink: ...

    @property
    def gcs_sink(self) -> GcsSink: ...

    def __init__(self,
        *,
        version : typing___Optional[typing___Text] = None,
        kind : typing___Optional[typing___Text] = None,
        name : typing___Optional[typing___Text] = None,
        bigquery_source : typing___Optional[BigQuerySource] = None,
        gcs_source : typing___Optional[GcsSource] = None,
        model : typing___Optional[Model] = None,
        bigquery_sink : typing___Optional[BigQuerySink] = None,
        gcs_sink : typing___Optional[GcsSink] = None,
        ) -> None: ...
    if sys.version_info >= (3,):
        @classmethod
        def FromString(cls, s: builtin___bytes) -> PredictionJob: ...
    else:
        @classmethod
        def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> PredictionJob: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"bigquery_sink",b"bigquery_sink",u"bigquery_source",b"bigquery_source",u"gcs_sink",b"gcs_sink",u"gcs_source",b"gcs_source",u"model",b"model",u"sink",b"sink",u"source",b"source"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"bigquery_sink",b"bigquery_sink",u"bigquery_source",b"bigquery_source",u"gcs_sink",b"gcs_sink",u"gcs_source",b"gcs_source",u"kind",b"kind",u"model",b"model",u"name",b"name",u"sink",b"sink",u"source",b"source",u"version",b"version"]) -> None: ...
    @typing___overload
    def WhichOneof(self, oneof_group: typing_extensions___Literal[u"sink",b"sink"]) -> typing_extensions___Literal["bigquery_sink","gcs_sink"]: ...
    @typing___overload
    def WhichOneof(self, oneof_group: typing_extensions___Literal[u"source",b"source"]) -> typing_extensions___Literal["bigquery_source","gcs_source"]: ...

class BigQuerySource(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class OptionsEntry(google___protobuf___message___Message):
        DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
        key = ... # type: typing___Text
        value = ... # type: typing___Text

        def __init__(self,
            *,
            key : typing___Optional[typing___Text] = None,
            value : typing___Optional[typing___Text] = None,
            ) -> None: ...
        if sys.version_info >= (3,):
            @classmethod
            def FromString(cls, s: builtin___bytes) -> BigQuerySource.OptionsEntry: ...
        else:
            @classmethod
            def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> BigQuerySource.OptionsEntry: ...
        def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
        def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"key",b"key",u"value",b"value"]) -> None: ...

    table = ... # type: typing___Text
    features = ... # type: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text]

    @property
    def options(self) -> typing___MutableMapping[typing___Text, typing___Text]: ...

    def __init__(self,
        *,
        table : typing___Optional[typing___Text] = None,
        features : typing___Optional[typing___Iterable[typing___Text]] = None,
        options : typing___Optional[typing___Mapping[typing___Text, typing___Text]] = None,
        ) -> None: ...
    if sys.version_info >= (3,):
        @classmethod
        def FromString(cls, s: builtin___bytes) -> BigQuerySource: ...
    else:
        @classmethod
        def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> BigQuerySource: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"features",b"features",u"options",b"options",u"table",b"table"]) -> None: ...

class GcsSource(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class OptionsEntry(google___protobuf___message___Message):
        DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
        key = ... # type: typing___Text
        value = ... # type: typing___Text

        def __init__(self,
            *,
            key : typing___Optional[typing___Text] = None,
            value : typing___Optional[typing___Text] = None,
            ) -> None: ...
        if sys.version_info >= (3,):
            @classmethod
            def FromString(cls, s: builtin___bytes) -> GcsSource.OptionsEntry: ...
        else:
            @classmethod
            def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> GcsSource.OptionsEntry: ...
        def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
        def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"key",b"key",u"value",b"value"]) -> None: ...

    format = ... # type: FileFormat
    uri = ... # type: typing___Text
    features = ... # type: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text]

    @property
    def options(self) -> typing___MutableMapping[typing___Text, typing___Text]: ...

    def __init__(self,
        *,
        format : typing___Optional[FileFormat] = None,
        uri : typing___Optional[typing___Text] = None,
        features : typing___Optional[typing___Iterable[typing___Text]] = None,
        options : typing___Optional[typing___Mapping[typing___Text, typing___Text]] = None,
        ) -> None: ...
    if sys.version_info >= (3,):
        @classmethod
        def FromString(cls, s: builtin___bytes) -> GcsSource: ...
    else:
        @classmethod
        def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> GcsSource: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"features",b"features",u"format",b"format",u"options",b"options",u"uri",b"uri"]) -> None: ...

class Model(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class OptionsEntry(google___protobuf___message___Message):
        DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
        key = ... # type: typing___Text
        value = ... # type: typing___Text

        def __init__(self,
            *,
            key : typing___Optional[typing___Text] = None,
            value : typing___Optional[typing___Text] = None,
            ) -> None: ...
        if sys.version_info >= (3,):
            @classmethod
            def FromString(cls, s: builtin___bytes) -> Model.OptionsEntry: ...
        else:
            @classmethod
            def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> Model.OptionsEntry: ...
        def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
        def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"key",b"key",u"value",b"value"]) -> None: ...

    class ModelResult(google___protobuf___message___Message):
        DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
        type = ... # type: ResultType
        item_type = ... # type: ResultType

        def __init__(self,
            *,
            type : typing___Optional[ResultType] = None,
            item_type : typing___Optional[ResultType] = None,
            ) -> None: ...
        if sys.version_info >= (3,):
            @classmethod
            def FromString(cls, s: builtin___bytes) -> Model.ModelResult: ...
        else:
            @classmethod
            def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> Model.ModelResult: ...
        def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
        def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"item_type",b"item_type",u"type",b"type"]) -> None: ...

    type = ... # type: ModelType
    uri = ... # type: typing___Text

    @property
    def result(self) -> Model.ModelResult: ...

    @property
    def options(self) -> typing___MutableMapping[typing___Text, typing___Text]: ...

    def __init__(self,
        *,
        type : typing___Optional[ModelType] = None,
        uri : typing___Optional[typing___Text] = None,
        result : typing___Optional[Model.ModelResult] = None,
        options : typing___Optional[typing___Mapping[typing___Text, typing___Text]] = None,
        ) -> None: ...
    if sys.version_info >= (3,):
        @classmethod
        def FromString(cls, s: builtin___bytes) -> Model: ...
    else:
        @classmethod
        def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> Model: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def HasField(self, field_name: typing_extensions___Literal[u"result",b"result"]) -> builtin___bool: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"options",b"options",u"result",b"result",u"type",b"type",u"uri",b"uri"]) -> None: ...

class BigQuerySink(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class OptionsEntry(google___protobuf___message___Message):
        DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
        key = ... # type: typing___Text
        value = ... # type: typing___Text

        def __init__(self,
            *,
            key : typing___Optional[typing___Text] = None,
            value : typing___Optional[typing___Text] = None,
            ) -> None: ...
        if sys.version_info >= (3,):
            @classmethod
            def FromString(cls, s: builtin___bytes) -> BigQuerySink.OptionsEntry: ...
        else:
            @classmethod
            def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> BigQuerySink.OptionsEntry: ...
        def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
        def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"key",b"key",u"value",b"value"]) -> None: ...

    table = ... # type: typing___Text
    staging_bucket = ... # type: typing___Text
    result_column = ... # type: typing___Text
    save_mode = ... # type: SaveMode

    @property
    def options(self) -> typing___MutableMapping[typing___Text, typing___Text]: ...

    def __init__(self,
        *,
        table : typing___Optional[typing___Text] = None,
        staging_bucket : typing___Optional[typing___Text] = None,
        result_column : typing___Optional[typing___Text] = None,
        save_mode : typing___Optional[SaveMode] = None,
        options : typing___Optional[typing___Mapping[typing___Text, typing___Text]] = None,
        ) -> None: ...
    if sys.version_info >= (3,):
        @classmethod
        def FromString(cls, s: builtin___bytes) -> BigQuerySink: ...
    else:
        @classmethod
        def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> BigQuerySink: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"options",b"options",u"result_column",b"result_column",u"save_mode",b"save_mode",u"staging_bucket",b"staging_bucket",u"table",b"table"]) -> None: ...

class GcsSink(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class OptionsEntry(google___protobuf___message___Message):
        DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
        key = ... # type: typing___Text
        value = ... # type: typing___Text

        def __init__(self,
            *,
            key : typing___Optional[typing___Text] = None,
            value : typing___Optional[typing___Text] = None,
            ) -> None: ...
        if sys.version_info >= (3,):
            @classmethod
            def FromString(cls, s: builtin___bytes) -> GcsSink.OptionsEntry: ...
        else:
            @classmethod
            def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> GcsSink.OptionsEntry: ...
        def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
        def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"key",b"key",u"value",b"value"]) -> None: ...

    format = ... # type: FileFormat
    uri = ... # type: typing___Text
    result_column = ... # type: typing___Text
    save_mode = ... # type: SaveMode

    @property
    def options(self) -> typing___MutableMapping[typing___Text, typing___Text]: ...

    def __init__(self,
        *,
        format : typing___Optional[FileFormat] = None,
        uri : typing___Optional[typing___Text] = None,
        result_column : typing___Optional[typing___Text] = None,
        save_mode : typing___Optional[SaveMode] = None,
        options : typing___Optional[typing___Mapping[typing___Text, typing___Text]] = None,
        ) -> None: ...
    if sys.version_info >= (3,):
        @classmethod
        def FromString(cls, s: builtin___bytes) -> GcsSink: ...
    else:
        @classmethod
        def FromString(cls, s: typing___Union[builtin___bytes, builtin___buffer, builtin___unicode]) -> GcsSink: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def ClearField(self, field_name: typing_extensions___Literal[u"format",b"format",u"options",b"options",u"result_column",b"result_column",u"save_mode",b"save_mode",u"uri",b"uri"]) -> None: ...



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Iterable, List

from pyspark.sql import DataFrame, SparkSession

from merlinpyspark.config import BigQuerySourceConfig, SourceConfig


def create_source(spark_session: SparkSession,
                  source_config: SourceConfig) -> "Source":
    if source_config.source_type() == BigQuerySourceConfig.TYPE:
        if not isinstance(source_config, BigQuerySourceConfig):
            raise ValueError('source_config is not BigQuerySourceConfig '
                             'instance')

        return BigQuerySource(spark_session, source_config)

    raise NotImplementedError(f"{source_config.source_type()} source type "
                              f"is not implemented")


class Source(ABC):
    @abstractmethod
    def load(self) -> DataFrame:
        """
        Loads a data stream from a data source and returns it as a :class`DataFrame`.
        :return: Spark DataFrame
        """
        pass

    @abstractmethod
    def features(self) -> Iterable[str]:
        """
        List of columns used for prediction
        :return: list of string
        """
        pass


class BigQuerySource(Source):
    READ_FORMAT = "bigquery"
    OPTION_TABLE = "table"
    OPTION_PARALLELISM = "maxParallelism"

    def __init__(self, spark_session: SparkSession,
                 bq_source_config: BigQuerySourceConfig):
        self._spark = spark_session
        self._config = bq_source_config

        if bq_source_config.table() is None:
            raise ValueError("table field is empty")

    def load(self) -> DataFrame:
        cfg = self._config
        features = self.features()

        reader = self._spark.read.format(self.READ_FORMAT) \
            .option(self.OPTION_TABLE, cfg.table())

        if cfg.options() is not None:
            reader.options(**cfg.options())

        df = reader.load()
        if features is not None:
            return df.select(*features)

        return df

    def features(self) -> Iterable[str]:
        return self._config.features()



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from test.model_test import IrisModel

import mlflow
import pytest
from merlinpyspark.__main__ import start
from merlinpyspark.config import load
from mlflow.pyfunc import log_model

test_config = """
kind: PredictionJob
version: v1
name: integration-test
bigquerySource:
  table: 'project.dataset.table_iris'
  features:
    - sepal_length
    - sepal_width
    - petal_length
    - petal_width
model:
  type: PYFUNC_V2
  uri: {}
  result:
    type: DOUBLE
bigquerySink:
  table: 'project.dataset.table_iris_result'
  result_column: 'prediction'
  save_mode: OVERWRITE
  options:
    project: "project"
    temporaryGcsBucket: "bucket-name"
"""


@pytest.mark.ci
def test_integration(spark_session, bq):
    log_model(
        "model",
        python_model=IrisModel(),
        artifacts={"model_path": "test-model/artifacts/model.joblib"},
    )
    model_path = os.path.join(mlflow.get_artifact_uri(), "model")

    config_path = "test-config/integration_test.yaml"
    with open(config_path, "w") as f:
        f.write(test_config.format(model_path))

    cfg = load(config_path)

    bq.delete_table(table=cfg.sink().table(), not_found_ok=True)

    start(config_path, spark_session)

    result_table = bq.get_table(cfg.sink().table())
    assert len(result_table.schema) == 5
    assert result_table.schema[4].name == cfg.sink().result_column()
    assert result_table.schema[4].field_type == "FLOAT"



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil

import pytest
from google.cloud import bigquery
from pyspark import SparkConf, SparkContext
from pyspark.sql import SparkSession

shutil.rmtree("mlruns", ignore_errors=True)

@pytest.fixture(scope="session")
def spark_session(request):
    conf = SparkConf()
    conf.set("spark.jars",
             "https://storage.googleapis.com/hadoop-lib/gcs/gcs-connector"
             "-hadoop2-2.0.1.jar")
    conf.set("spark.jars.packages",
             "com.google.cloud.spark:spark-bigquery-with-dependencies_2.11:0.13.1-beta")
    sc = SparkContext(conf=conf)

    sc._jsc.hadoopConfiguration().set("fs.gs.impl",
                                      "com.google.cloud.hadoop.fs.gcs.GoogleHadoopFileSystem")
    sc._jsc.hadoopConfiguration().set("fs.AbstractFileSystem.gs.impl",
                                      "com.google.cloud.hadoop.fs.gcs.GoogleHadoopFS")
    sc._jsc.hadoopConfiguration().set(
        "google.cloud.auth.service.account.enable", "true")

    sa_path = os.environ.get("GOOGLE_APPLICATION_CREDENTIALS")
    if sa_path is not None:
        sc._jsc.hadoopConfiguration().set(
            "google.cloud.auth.service.account.json.keyfile",
            sa_path)

    spark = SparkSession.builder \
        .config(conf=sc.getConf()) \
        .getOrCreate()

    request.addfinalizer(lambda: spark.stop())

    return spark


@pytest.fixture
def bq():
    return bigquery.Client()



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from merlinpyspark.config import BigQuerySourceConfig
from merlinpyspark.source import create_source
from merlinpyspark.spec.prediction_job_pb2 import BigQuerySource
import pytest

@pytest.mark.ci
def test_bq_source(spark_session):
    bq_src_proto = BigQuerySource(
        table="bigquery-public-data:samples.shakespeare",
        features=["word", "word_count"],
        options={"filter": "word_count > 100"}
    )
    src_cfg = BigQuerySourceConfig(bq_src_proto)

    bq_source = create_source(spark_session, src_cfg)
    df = bq_source.load()
    assert len(df.columns) == 2
    assert df.columns[0] == "word"
    assert df.columns[1] == "word_count"
    assert df.sort("word_count").take(1)[0]["word_count"] > 100
